// SPDX-License-Identifier: MIT

use std::{fs, io, os::unix::io::AsRawFd, path::Path};

use super::Hid;
use crate::Error;

pub struct HidRaw(pub fs::File);

impl HidRaw {
    pub fn open<P: AsRef<Path>>(path: P) -> Result<Self, Error> {
        Ok(Self(fs::File::options().read(true).write(true).open(path)?))
    }
}

impl Hid for HidRaw {
    fn write(&mut self, buf: &[u8]) -> Result<usize, Error> {
        use nix::unistd::*;
        Ok(write(self.0.as_raw_fd(), buf).map_err(io::Error::from)?)
    }

    // Based on `hid_read_timeout()` in hidapi
    fn read_timeout(&mut self, buf: &mut [u8], timeout: i32) -> Result<usize, Error> {
        use nix::{poll::*, unistd::*};
        if timeout >= 0 {
            let pollfd = PollFd::new(self.0.as_raw_fd(), PollFlags::POLLIN);
            let res = poll(&mut [pollfd], timeout).map_err(io::Error::from)?;
            if res == 0 {
                // Timeout
                return Ok(0);
            }
            if let Some(revents) = pollfd.revents() {
                if revents.intersects(PollFlags::POLLERR | PollFlags::POLLHUP | PollFlags::POLLNVAL)
                {
                    return Err(io::Error::new(
                        io::ErrorKind::Other,
                        format!("poll error: {:?}", revents),
                    )
                    .into());
                }
            }
        }
        Ok(read(self.0.as_raw_fd(), buf).map_err(io::Error::from)?)
    }
}
